/// <reference types="node" />
export type generate_uuid = () => string;
export type uuid4 = () => Buffer;
export type uuid_to_string = (buffer: Buffer) => string;
export type string_to_uuid = (uuid_string: string) => Buffer;
export type clone = (o: any) => any;
export type and = (f: Function, g: Function) => Function;
export type is_sender = (o: any) => boolean;
export type is_receiver = (o: any) => boolean;
export type sender_filter = (filter: any) => Function;
export type receiver_filter = (filter: any) => Function;
export type is_defined = (field: any) => boolean;
